package be.iminds.ilabt.jfed.restapi.client;

import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClientDirect;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.FedmonApiCommon;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.JsonLdObjects;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection.BasicConnectionBuilderFactory;
import be.iminds.ilabt.jfed.lowlevel.connection.ConnectionBuilder;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.restapi.json.Experiment;
import be.iminds.ilabt.jfed.restapi.json.ExperimentBuilder;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredential;
import be.iminds.ilabt.jfed.util.IOUtils;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.io.ByteStreams;
import io.dropwizard.jackson.Jackson;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.*;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.net.ssl.SSLException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.*;

/**
 * JFedRestApiClientImpl implementation of JFedRestApiClient interface using apache HttpClient
 */
public class JFedRestApiClientImpl implements JFedRestApiClient {
    private static final Logger LOG = LoggerFactory.getLogger(JFedRestApiClientImpl.class);

    private final JFedRestApiClientConfig config;
    private final BasicConnectionBuilderFactory connectionBuilderFactory  = new BasicConnectionBuilderFactory();

    public JFedRestApiClientImpl(JFedRestApiClientConfig config) {
        this.config = config;
    }

    @Override
    public Optional<Experiment> getById(Class<? extends Experiment> objectClass, Long id, boolean full) throws JFedRestApiClientException {
        URIBuilder builder = config.getUriBuilder(Experiment.class, ""+id, true/*read*/);
        if (full)
            builder.addParameter("full", "true");

        ServerResponse serverResponse = contactServer(builder, null/*jsonData*/, HttpMethod.GET);
        serverResponse.throwIfNotOk();

        return processOptionalResultHelper(serverResponse, Experiment.class);
    }

    @Override
    public Map<Long, Experiment> getAllIndexedById(Class<? extends Experiment> objectClass) throws JFedRestApiClientException {
        throw new RuntimeException("not yet implemented");
    }

    @Override
    public List<Experiment> getAll(Class<? extends Experiment> objectClass) throws JFedRestApiClientException {
        throw new RuntimeException("not yet implemented");
    }

    @Override
    public Experiment create(Experiment object) throws JFedRestApiClientException {
        Class<Experiment> objectClass = Experiment.class;
        String jsonData;
        try {
            jsonData = MAPPER.writer().writeValueAsString(object);
        } catch (JsonProcessingException e) {
            throw new JFedRestApiClientException("Error converting "+objectClass.getSimpleName()+" to JSON", e);
        }

        ServerResponse serverResponse = contactServer(
                config.getUriBuilder(objectClass, false/*read*/),
                jsonData, HttpMethod.POST);

        serverResponse.throwIfNotOk();
        serverResponse.throwIfNoData();

        Experiment reply = processResultHelper(serverResponse, objectClass);
        if (reply == null)
            throw new JFedRestApiClientException("Got no reply on POST "+objectClass.getSimpleName());
        assert reply.getId() != null;

        return reply;
    }

    @Override
    public Experiment update(Experiment object) throws JFedRestApiClientException {
        assert object != null;
        assert object.getId() != null;
        Class<Experiment> objectClass = Experiment.class;

        ExperimentBuilder builder = new ExperimentBuilder(object);

        //TODO: minimize based on annotation instead
        Experiment minimized;
        if (objectClass.equals(Testbed.class))
            minimized = builder.createMinimized(JsonLdObjects.Minimization.FULL_NO_CHILDREN_PARENT_ID_ONLY);
        else
            minimized = builder.createMinimized(JsonLdObjects.Minimization.FULL_WITH_LINK_IDS);

        String jsonData;
        try {
            jsonData = MAPPER.writer().writeValueAsString(minimized);
        } catch (JsonProcessingException e) {
            throw new JFedRestApiClientException("Error converting "+objectClass.getSimpleName()+" to JSON", e);
        }

        ServerResponse serverResponse = contactServer(
                config.getUriBuilder(objectClass, object.getId()+"", false/*read*/),
                jsonData, HttpMethod.PUT);

        serverResponse.throwIfNotOk();
        serverResponse.throwIfNoData();

        return processResultHelper(serverResponse, objectClass);
    }

    @Override
    public Experiment delete(Experiment object) throws JFedRestApiClientException {
        assert object != null;
        assert object.getId() != null;

        return deleteById(object.getClass(), object.getId());
    }

    @Override
    public Experiment deleteById(Class<? extends Experiment> objectClass, Long id) throws JFedRestApiClientException {
        URIBuilder builder = config.getUriBuilder(objectClass, ""+id, false/*read*/);
        ServerResponse serverResponse = contactServer(builder, null/*jsonData*/, HttpMethod.DELETE);

        //should return nothing
        serverResponse.throwIfNotOk();
        serverResponse.throwIfNoData();

        return processResultHelper(serverResponse, objectClass);
    }

    @Override
    public Optional<SpeaksForCredential> getSpeaksForCredential(String tld, String username) throws JFedRestApiClientException {
        URIBuilder builder = config.getUriBuilder(Experiment.class, ""+tld+"/"+username, true/*read*/);

        ServerResponse serverResponse = contactServer(builder, null/*jsonData*/, HttpMethod.GET);
        serverResponse.throwIfNotOk();

        return processOptionalResultHelper(serverResponse, SpeaksForCredential.class);
    }

    @Override
    public void upsertSpeaksForCredential(SpeaksForCredential object) throws JFedRestApiClientException {
        assert object.getSpeakingFor() != null;
//        String tld = object.getSpeakingFor().getEncodedTopLevelAuthority();
//        String username = object.getSpeakingFor().getEncodedResourceName();

        assert object.getCredentialXml() != null;

        upsertSpeaksForCredential(/*tld, username, */object.getCredentialXml());
    }
    @Override
    public void upsertSpeaksForCredential(@Nonnull String speaksForCredentialXmlString) throws JFedRestApiClientException {
        Class<SpeaksForCredential> objectClass = SpeaksForCredential.class;

        StringEntity myEntity = new StringEntity(speaksForCredentialXmlString, ContentType.create("application/xml", "UTF-8"));

        ServerResponse serverResponse = null;
        try {
            serverResponse = contactServer(
                    config.getUriBuilder(objectClass, ""/*+tld+"/"+username*/, false/*read*/).build(),
                    myEntity,
                    HttpMethod.POST,
                    null/*rsponseHandler*/,
                    null/*alternateAcceptHeader*/);
        } catch (URISyntaxException e) {
            throw new JFedRestApiClientException("Error constructing URI", e);
        }

        serverResponse.throwIfNotOk();
        serverResponse.throwIfNoData();

        assert serverResponse.getHttpCode() == 201; //Created
        String uri = serverResponse.getJsonReply();
        LOG.debug("upsertSpeaksForCredential reply: \""+uri+"\"");
//
//        SpeaksForCredential reply = processResultHelper(serverResponse, objectClass);
//        if (reply == null)
//            throw new JFedRestApiClientException("Got no reply on POST "+objectClass.getSimpleName());
//        assert reply.getUri() != null;
//
//        return reply;
    }

    @Override
    public void deleteSpeaksForCredential(SpeaksForCredential object) throws JFedRestApiClientException {
        assert object != null;
        Class<SpeaksForCredential> objectClass = SpeaksForCredential.class;

        assert object.getSpeakingFor() != null;
        String tld = object.getSpeakingFor().getEncodedTopLevelAuthority();
        String username = object.getSpeakingFor().getEncodedResourceName();

        URIBuilder builder = config.getUriBuilder(objectClass, ""+tld+"/"+username, false/*read*/);
        ServerResponse serverResponse = contactServer(builder, null/*jsonData*/, HttpMethod.DELETE);

        //should return nothing
        serverResponse.throwIfNotOk();
        serverResponse.throwIfData();
    }

    @Override
    public void deleteByUri(URI uri) throws JFedRestApiClientException {
        ServerResponse serverResponse = contactServer(uri, (HttpEntity) null/*jsonData*/, HttpMethod.DELETE, null, null);

        //should return nothing
        serverResponse.throwIfNotOk();
//        serverResponse.throwIfNoData(); //Experiment returns data, others do not

//        return processResultHelper(serverResponse, Experiment.class);
    }


    private static final ObjectMapper MAPPER;
    static {
        MAPPER = Jackson.newObjectMapper();
        //the client should not be strict on unknown properties
        MAPPER.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

    private static class ServerResponse {
        private final int httpCode;
        private final String origUrlAsString;
        private final String jsonReplyOrErrorMessage;
        private final boolean noData;

        public ServerResponse(int httpCode, boolean noData, String jsonReplyOrErrorMessage) {
            this.httpCode = httpCode;
            this.jsonReplyOrErrorMessage = jsonReplyOrErrorMessage;
            this.noData = noData;
            this.origUrlAsString = null;
        }
        public ServerResponse(int httpCode, boolean noData, String jsonReplyOrErrorMessage, String origUrlAsString) {
            this.httpCode = httpCode;
            this.jsonReplyOrErrorMessage = jsonReplyOrErrorMessage;
            this.noData = noData;
            this.origUrlAsString = origUrlAsString;
        }
        public ServerResponse(int httpCode, boolean noData, String jsonReplyOrErrorMessage, URI origUrl) {
            this.httpCode = httpCode;
            this.jsonReplyOrErrorMessage = jsonReplyOrErrorMessage;
            this.noData = noData;
            this.origUrlAsString = origUrl == null ? null : origUrl.toASCIIString();
        }

        public int getHttpCode() {
            return httpCode;
        }
        public boolean is404() {
            return httpCode == 404;
        }
        public boolean isOk() {
            return httpCode >= 200 && httpCode < 300; //TODO check if these are all acceptable codes
        }

        public boolean hasData() {
            return !noData;
        }
        public boolean hasNoData() {
            return noData;
        }

        public String getJsonReply() {
            return jsonReplyOrErrorMessage;
        }

        public void throwIfData() throws JFedRestApiClientException {
            if (hasData())
                throw new JFedRestApiClientException("HTTP request unexpectedly returned data"+urlHelp());
        }
        public void throwIfNoData() throws JFedRestApiClientException {
            if (hasNoData())
                throw new JFedRestApiClientException("HTTP request unexpectedly returned no data."+urlHelp()+
                        " code="+httpCode+(jsonReplyOrErrorMessage == null ? "" : " msg="+jsonReplyOrErrorMessage));
        }
        public void throwIf404() throws JFedRestApiClientException {
            if (is404())
                throw new JFedRestApiClientException("HTTP request returned unexpected 404 ("+httpCode+")"+urlHelp()+
                        " code="+httpCode+(jsonReplyOrErrorMessage == null ? "" : " msg="+jsonReplyOrErrorMessage));
        }
        public void throwIfNotOk() throws JFedRestApiClientException {
            if (!isOk())
                throw new JFedRestApiClientException("HTTP request returned unexpected code."+urlHelp()+
                        " code="+httpCode+(jsonReplyOrErrorMessage != null ? (" message="+ jsonReplyOrErrorMessage) : ""));
        }
        public void throwIfNotOkOr404() throws JFedRestApiClientException {
            if (!is404() && !isOk())
                throw new JFedRestApiClientException("HTTP request returned unexpected code. "+urlHelp()+
                        " code="+httpCode+(jsonReplyOrErrorMessage == null ? "" : " msg="+jsonReplyOrErrorMessage));
        }
        private String urlHelp() {
            if (origUrlAsString == null || origUrlAsString.trim().isEmpty())
                return "";
            else
                return " url=\""+origUrlAsString+"\"";
        }
    }

    private static enum HttpMethod {GET, DELETE, POST, PUT}

    private ServerResponse contactServer(URIBuilder builder, String jsonData, HttpMethod httpMethod) throws JFedRestApiClientException {
        try {
            URI uri = builder.build();
            return contactServer(uri, jsonData, httpMethod, null, null);
        } catch (URISyntaxException e) {
            throw new JFedRestApiClientException("Error constructing URI", e);
        }
    }


    private static interface ResponseHandler {
        public ServerResponse handleOkResponse(HttpResponse response);
    }
    private static class BinaryResponseHandler implements ResponseHandler {
        private byte[] data = null;

        @Override
        public ServerResponse handleOkResponse(HttpResponse response) {
            assert response != null;

            HttpEntity entity = response.getEntity();
            if (entity != null) {
                try (InputStream replyContent = entity.getContent()) {
                    data = ByteStreams.toByteArray(replyContent);
                } catch (IOException e) {
                    data = null;
                    LOG.error("Error reading server response", e);
                }
            } else {
                LOG.debug("No response entity -> no data");
                data = null;
            }

            //we lost the URI of the call, so we can't log it here
            return new ServerResponse(response.getStatusLine().getStatusCode(), data == null || data.length == 0/*noData*/, null, (URI)null/*origURL*/);
        }

        public byte[] getData() {
            return data;
        }
    }

    /**
     *
     * @param uri
     * @param jsonData a string containing the JSON data
     * @param httpMethod
     * @param alternateResponseHandler
     * @param alternateAcceptHeader
     * @return null if no return data. empty optional if 404. optional with result string otherwise.
     * @throws IOException
     */
    private ServerResponse contactServer(URI uri, String jsonData, HttpMethod httpMethod, ResponseHandler alternateResponseHandler, String alternateAcceptHeader) throws JFedRestApiClientException {
        StringEntity myEntity = jsonData == null ? null : new StringEntity(jsonData, ContentType.create("application/json", "UTF-8"));

        return contactServer(uri, myEntity, httpMethod, alternateResponseHandler, alternateAcceptHeader);
    }

    private ServerResponse contactServer(URI uri, HttpEntity dataEntity, HttpMethod httpMethod, ResponseHandler alternateResponseHandler, String alternateAcceptHeader) throws JFedRestApiClientException {

        ConnectionBuilder connectionBuilder = connectionBuilderFactory.createConnectionBuilder();
        HttpConnection con;
        try {
            connectionBuilder.setUrl(uri.toURL());

            if (uri.getScheme().equals("http")) {
                connectionBuilder.useHttp();
                connectionBuilder.useNoAuthentication();
            } else {
                if (uri.getScheme().equals("https")) {
                    connectionBuilder.useHttps(config.getTrustStore(), null/*handleUntrustedCallback*/);
                    if (config.getUserPrivateKey() != null &&
                            config.getUserCertificateChain() != null &&
                            !config.getUserCertificateChain().isEmpty()) {
                        connectionBuilder.useSslClientAuthentication(config.getUserCertificateChain(), config.getUserPrivateKey());
                    } else {
                        connectionBuilder.useNoAuthentication();
                    }
                } else
                    throw new JFedRestApiClientException("Unsupported scheme \""+uri.getScheme()+"\" in URI \""+uri.toASCIIString()+"\"");
            }

            //explicit mention of unused config
            connectionBuilder.setDebugInfo(new JFedConnection.DebugInfo(null/*service*/, null/*user*/));
            connectionBuilder.setHackAllowAllServerCertificates(false);
            connectionBuilder.setProxy(null/*proxyInfo*/, false/*forceSshProxyDns*/);

            con = connectionBuilder.buildHttpConnection();
            assert con != null;
        } catch (JFedException e) {
            throw new JFedRestApiClientException("Problem creating connection", e);
        } catch (MalformedURLException e) {
            throw new JFedRestApiClientException("MalformedURLException for \""+uri.toASCIIString()+"\"", e);
        }

        HttpClient httpClient = con.getHttpClient();

//        HttpParams params = new BasicHttpParams();
//        params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 120000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
//        params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)
//        DefaultHttpClient httpClient = new DefaultHttpClient(params);



//        HttpEntityEnclosingRequestBase httpSend;
        HttpRequestBase httpSend;
        switch (httpMethod) {
            case POST: {
                HttpPost httpPost = new HttpPost(uri);
                httpSend = httpPost;
                if (dataEntity != null)
                    httpPost.setEntity(dataEntity);
                break;
            }
            case PUT: {
                HttpPut httpPut = new HttpPut(uri);
                httpSend = httpPut;
                if (dataEntity != null)
                    httpPut.setEntity(dataEntity);
                break;
            }
            case DELETE: { httpSend = new HttpDelete(uri); break; }
            case GET: {
                httpSend = new HttpGet(uri);
                httpSend.addHeader("accept", alternateAcceptHeader != null ? alternateAcceptHeader : "application/json");
                break; }
            default:
                throw new JFedRestApiClientException("Operation not supported: " + httpMethod);
        }
        HttpResponse response = null;
        try {
            if (LOG.isDebugEnabled()) {
                String httpRequestLine = httpSend.getRequestLine().toString();

                StringBuilder httpRequestHeaders = new StringBuilder(200);
                for (Header header : httpSend.getAllHeaders())
                    httpRequestHeaders.append(header.toString() + "\n");
                httpRequestHeaders.append("\n");

                LOG.debug("contactServer(" + httpMethod + ") request: httpRequestLine=" + httpRequestLine);
                LOG.debug("contactServer(" + httpMethod + ") request: httpRequestHeaders=" + httpRequestHeaders);
            }

            if (LOG.isDebugEnabled()) {
                LOG.debug("contactServer(" + httpMethod + ") is sending " + (dataEntity == null ? "no content" : (dataEntity.getContentLength()+ " bytes")) );
            }
            con.markInUse();
            response = con.getHttpContext() != null ?
                    httpClient.execute(httpSend, con.getHttpContext()) :
                    httpClient.execute(httpSend);

            if (LOG.isDebugEnabled())
                LOG.debug("contactServer(" + httpMethod + ") has sent. response.getStatusLine()=" + response.getStatusLine());

            if (LOG.isDebugEnabled()) {
                StatusLine httpResultStatusLine = null;
                StringBuilder httpResponseHeaders = null;
                List<Header> rawHttpResponseHeaders = null;

                if (response != null && response.getStatusLine() != null)
                    httpResultStatusLine = response.getStatusLine();

                if (response != null && response.getAllHeaders() != null) {
                    httpResponseHeaders = new StringBuilder(100);
                    for (Header header : response.getAllHeaders())
                        httpResponseHeaders.append(header.toString() + "\n");
                    httpResponseHeaders.append("\n");

                    rawHttpResponseHeaders = new ArrayList<Header>(Arrays.asList(response.getAllHeaders()));
                }

                LOG.debug("contactServer(" + httpMethod + ") reply: httpResultStatusLine=" + httpResultStatusLine);
                LOG.debug("contactServer(" + httpMethod + ") reply: httpResponseHeaders=" + httpResponseHeaders);
            }

            if (response.getStatusLine() == null)
                throw new JFedRestApiClientException("HTTP error: no response StatusLine");
            if (response.getStatusLine() != null && (response.getStatusLine().getStatusCode() < 200 || response.getStatusLine().getStatusCode() >= 300)) {
                if (LOG.isDebugEnabled()) {
//                    LOG.debug("contactServer(" + httpMethod + ") has sent the following data:\n" + jsonData);
                    LOG.error("Call failed: Got HTTP " + response.getStatusLine().getStatusCode() + ": " + response.getStatusLine().getReasonPhrase());
                    //don't do this: it will consume the response...
//                    HttpEntity entity = response.getEntity();
//                    if (entity != null) {
//                        InputStream replyContent = entity.getContent();
//                        if (replyContent != null) {
//                            String replyString = null;
//                            try {
//                                replyString = new Scanner(replyContent, "UTF-8").useDelimiter("\\A").next();
//                                LOG.debug("reply:\n" + replyString);
//                            } catch (NoSuchElementException e) {
//                                LOG.debug("empty stream from response.getEntity()");
//                            }
//                        } else
//                            LOG.debug("no stream from response.getEntity()");
//                    } else
//                        LOG.debug("no reply");
                }

                String replyString = "";
                if (response != null) {
                    HttpEntity entity = response.getEntity();
                    if (entity != null) {
                        InputStream replyContent = entity.getContent();
                        if (replyContent != null) {
                            try {
                                replyString = IOUtils.streamToString(replyContent, "UTF-8");
                            } finally {
                                replyContent.close();
                            }
                            //replyString = new Scanner(replyContent, "UTF-8").useDelimiter("\\A").next();
                        }
                    }
                }

                return new ServerResponse(response.getStatusLine().getStatusCode(), true, response.getStatusLine().getReasonPhrase(), uri);
//                throw new FedmonWebApiClientException("Call failed: Got HTTP " + response.getStatusLine().getStatusCode() + ": " + response.getStatusLine().getReasonPhrase() +
//                        "\n" + replyString +
//                        "\n\nTried to send:\n" + jsonData);
            }
        } catch (javax.net.ssl.SSLPeerUnverifiedException ex) {
            LOG.error("SSLPeerUnverifiedException exception.\nWill also call LogDebugInfoCallback to retrieve connection debug info.", ex);

            throw new JFedRestApiClientException("The server certificate for '" + httpSend.getURI() + "' could not be verified. " +
                    "Error message \"" + ex.getMessage() + "\". " +
                    "Possible causes:\n" +
                    "  - The server's self-signed certificate is not in our trust store. \n" +
                    "  - The server's certificate \"CN\" field is not the server hostname or a known alias. \n" +
                    "  - The server is not accepting our client login certificate+key pair (perhaps is is not federated with our \"login provider\"). \n" +
                    "  - If the server's certificate is not self signed, we might not have the root certificate of the trust chain in our trust store.", ex);
        } catch (SSLException ex) {
            throw new JFedRestApiClientException("SSLException: " + ex.getMessage(), ex);
        } catch (ClientProtocolException ex) {
            throw new JFedRestApiClientException("ClientProtocolException: " + ex.getMessage(), ex);
        } catch (IOException ex) {
            throw new JFedRestApiClientException("IOException: " + ex.getMessage(), ex);
        } finally {
            con.markNotInUse();
        }

        if (response != null) {
            if (alternateResponseHandler != null) {
                //alternateResponseHandler is bit of a hack, could perhaps be implemented cleaner
                return alternateResponseHandler.handleOkResponse(response);
            } else {
                try {
                    String replyString = null;
                    if (response != null) {
                        HttpEntity entity = response.getEntity();
                        if (entity != null) {
                            InputStream replyContent = entity.getContent();
                            if (replyContent != null) {
                                try {
                                    replyString = IOUtils.streamToString(replyContent, "UTF-8");
                                    //replyString = new Scanner(replyContent, "UTF-8").useDelimiter("\\A").next();
                                } finally {
                                    try {
                                        replyContent.close();
                                    } catch (IOException e) {
                                        //ignore
                                    }
                                }
                            }
                        }
                    }
                    return new ServerResponse(response.getStatusLine().getStatusCode(), replyString == null, replyString, uri);
                } catch (IOException e) {
                    LOG.error("Error consuming reply content");
                }
            }
        }
        return null;
    }



    private static <T> T processResultHelper(ServerResponse serverResponse, Class<T> resClass) throws JFedRestApiClientException {
        if (serverResponse.is404())
            return null;
        try {
            T res = MAPPER.readValue(serverResponse.getJsonReply(), resClass);
            return res;
        } catch (IOException ex) {
            throw new JFedRestApiClientException("Json parse exception fetching "+resClass.getName()+". reply: "+serverResponse.getJsonReply(), ex);
        }
    }
    private static <T> Optional<T> processOptionalResultHelper(ServerResponse serverResponse, Class<T> resClass) throws JFedRestApiClientException {
        if (serverResponse.is404())
            return Optional.empty();
        try {
            String jsonReply = serverResponse.getJsonReply();
            assert jsonReply != null : "JSON reply is unexpectedly null. code="+serverResponse.getHttpCode()+" nodata="+serverResponse.hasNoData();
            T res = MAPPER.readValue(jsonReply, resClass);
            return Optional.of(res);
        } catch (IOException ex) {
            throw new JFedRestApiClientException("Json parse exception fetching "+resClass.getName()+". reply: "+serverResponse.getJsonReply(), ex);
        }
    }
    private static <T> List<T> processListResultHelper(ServerResponse serverResponse, Class<T> resClass) throws JFedRestApiClientException {
        try {
            List<T> res = MAPPER.readValue(serverResponse.getJsonReply(), MAPPER.getTypeFactory().constructCollectionType(List.class, resClass));
            return res;
        } catch (IOException ex) {
            throw new JFedRestApiClientException("Json parse exception fetching List of "+resClass.getName()+". reply: "+serverResponse.getJsonReply(), ex);
        }
    }
}
